"""Elasticsearch Mapping for DACC Repository"""
__author__ = "Jeremy Nelson"

MAP = {"mappings": {
    "mods": {
        "properties": {
            "abstract": {
                "type": "text"
            },
            "adminNote": {
                "type": "text"
            },
            "content_models": {
              "index": "false",
                       "type": "text"
            },
            "contributor": {
                "type": "text"
            },
            "creator": {
                "type": "text"
            },
            "datastreams": {
               "properties": {
                       "dsid": {
                           "index": "false",
                           "type": "text"
                       },
					   "label": {
                           "index": "false",
                           "type": "text"
                       },
                       "mimeType": {
                           "index": "false",
                           "type": "text"
                       }
                }
            },
            "dateCreated": {
               "index": "false",
                "type": "text"
            },
            "dateIssued": {
                "type": "text"
            },
            "datePublished": {
                "index": "false",
                "type": "text"
	    },
            "degreeGrantor": {
                "type": "text"
            },
            "degreeName": {
                "type": "text"
            },
            "degreeType": {
                "type": "text"
            },
            "digitalOrigin": {
                "type": "text"
            },
            "extent": {
                "type": "text"
            },
            "genre": {
                "index": "false",
                "type": "text"
            },
            "handle": {
                "type": "text"
            },
            "inCollections": {
                "index": "false",
                "type": "text"
            },
            "language": {
                "type": "text",
                "fields": {
                    "keyword": {
                        "type":"keyword"
                    }
                }
            },
            "note": {
                "type": "text"
            },
            "parent": {
                "index": "false",
                "type": "text"
            },
            "photographer": {
                "type": "text"
            },

            "pid": {
                "index": "false",
                "type": "text"
            },
            "place": {
                "type": "text"
            },
            "publisher": {
                "type": "text"
            },
            "publicationYear": {             
                "index": "false",
                "type": "text"
            },
            "sponsor": {
                "type": "text"
            },
            "subject": {
                "properties": {
                    "genre": {
			"index": "false",
                        "type": "text"
                    },
                    "geographic": {
		        "index": "false",
                        "type": "text"
                    },
                    "temporal": {
		        "index": "false",
                        "type": "text"
                    },
                    "topic": {
			"index": "false",
                        "type": "text"
                    }
                }
            },
            "thesis": {
                "type": "text"
            },
            "thesisAdvisor": {
                "type": "text"
            },
            "titleAlternative": {
                "type": "text"
            },
            "titlePrincipal": {
                "type": "text",
                "fields": {
                    "keyword": {
                        "type": "keyword"
                    }
                }

            },
            "titleRaw": {
                "type": "text",
                "index": "false"
            },
            "typeOfResource": {
                "index": "false",
                "type": "text"
            },
            "useAndReproduction": {
                "type": "text"
            }
            }
        }
    }
}



"""
 :mod:`helpers` Fedora Batch App Helpers
"""
__author__ = "Jeremy Nelson"

import bibcat
import datetime
from . import forms
import json
import mimetypes
import os
import requests
import shutil
import urllib.parse
import xml.etree.ElementTree as etree
import xml.dom.minidom
from flask import Response, request
from jinja2 import Environment, FileSystemLoader
from types import SimpleNamespace

JINJA_ENV = Environment(loader=FileSystemLoader(
    os.path.dirname(os.path.abspath(__file__))))

DEFAULT_NS = {
    "fedora_manage": "http://www.fedora.info/definitions/1/0/management/",
    "mods": 'http://www.loc.gov/mods/v3',
    "dc": "http://purl.org/dc/elements/1.1/", 
    "fedora": "info:fedora/fedora-system:def/relations-external#",
    "fedora-model": "info:fedora/fedora-system:def/model#",
    "islandora": "http://islandora.ca/ontology/relsext#", 
    "oai_dc": "http://www.openarchives.org/OAI/2.0/oai_dc/", 
    "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#", 
    "rdfs": "http://www.w3.org/2000/01/rdf-schema#"
}
 
for key, value in DEFAULT_NS.items():
    etree.register_namespace(key, value)

def __create_language__(mods, lang_str):
    if len(lang_str) < 1:
        return
    lang = etree.SubElement(mods,
        "mods:language")
    lang.text = lang_str

def __create_name__(mods, name_str, type_, role_str=None):
    if len(name_str) < 1:
        return
    name = etree.SubElement(mods, 
         "mods:name", 
         attrib={"type": type_})
    name_part = etree.SubElement(name, "mods:namePart")
    name_part.text = name_str
    role = etree.SubElement(name, "mods:role")
    role_term = etree.SubElement(role,
        "mods:roleTerm",
        attrib={"type": "text",
                "authority": "marcrelator"})
    role_term.text = role_str

def __create_origin_info__(form, mods):
    origin_info = etree.SubElement(mods, "mods:originInfo")
    if len(form.date_created.data) > 0:
        date_created = etree.SubElement(origin_info,
            "mods:dateCreated",
            attrib={"keyDate": "yes"})
        date_created.text = form.date_created.data
    if len(form.publisher.data) > 0:
        publisher = etree.SubElement(origin_info,
            "mods:publisher")
        publisher.text = form.publisher.data
    if len(form.publication_place.data) > 0:
        place = etree.SubElement(origin_info,
            "mods:place")
        place_term = etree.SubElement(place,
            "mods:placeTerm")
        place_term.text = form.publication_place.data
    if len(form.frequency.data) > 0:
        if not form.frequency.data.startswith("choose"):
            frequency = etree.SubElement(origin_info,
                "mods:frequency")
            frequency.text = form.frequency.data

def __create_phy_desc__(form, mods):
    phys_desc = etree.SubElement(mods, "mods:physicalDescription")
    if len(form.form.data) > 0:
        form_bib = etree.SubElement(phys_desc,
            "mods:form")
        form_bib = form.form.data
    if len(form.digital_origin.data) > 0:
        dig_origin = etree.SubElement(phys_desc, "mods:digitalOrigin")
        dig_origin.text = form.digital_origin.data
    if len(form.extent.data) > 0:
        extent = etree.SubElement(phys_desc, "mods:extent")
        extent.text = form.extent.data
    if len(form.description.data) > 0:
        note = etree.SubElement(phys_desc, "mods:note")
        note.text = form.description.data
    

def __create_subjects__(form, mods):
    for row in [(form.subject_dates.data, "mods:temporal"),
                (form.subject_places.data, "mods:geographic"),
                (form.subject_topics.data, "mods:topic")]:
        for data in row[0]:
            if len(data) < 1:
                continue
            subject = etree.SubElement(mods,
                "mods:subject")
            child = etree.SubElement(subject,
                row[1])
            child.text = data
    for row in [(form.subject_orgs.data, "corporate"),
                (form.subject_people.data, "personal")]:
        for data in row[0]:
            subject = etree.SubElement(mods,
                "mods:subject")
            name = etree.SubElement(subject,
                "mods:name", 
                attrib={"type": row[1]})
            namePart = etree.SubElement(name,
                "mods:namePart")
            namePart.text = data

       

def __create_title__(mods, form, alt=False):
    title_info = etree.SubElement(mods,
        "mods:titleInfo")
    title = etree.SubElement(title_info,
        "mods:title")
    main_title = form.title
    if alt is True:
        alt_title = form.alt_title
        title_info.attrib["type"] = "alternative"
        title.text = alt_title.data
    else: 
        title.text = main_title.data
    if "subtitle" in form and form.subtitle is not None:
        sub_title = etree.SubElement(title_info,
            "mods:subTitle")
        sub_title.text = subtitle

def __create_type_of_resource__(mods, type_of):
    type_of_resource = etree.SubElement(mods,
        "mods:typeOfResource")
    type_of_resource.text = type_of
    

def build_mods(form, pid=None):
    """Builds a MODS etree populated on WTform values"""
    mods = etree.XML(
        """<mods:mods xmlns:mods="http://www.loc.gov/mods/v3" />""")
    for row in form.admin_notes.data:
        if len(row) < 1:
            continue
        note = etree.SubElement(mods, 
            "mods:note", 
            attrib={"type": "admin"})
        note.text = row
    if len(form.abstract.data) > 0:
        abstract = etree.SubElement(mods, "mods:abstract")
        abstract.text = form.abstract.data
    if len(form.genre.data) > 0 and not form.genre.data.startswith("choose"):
        genre = etree.SubElement(mods, "mods:genre")
        genre.text = form.genre.data
    for row in form.type_of_resources.data:
        __create_type_of_resource__(mods, row)
    if len(form.alt_title.data) > 0:
        __create_title__(mods, form, True)
    for row in form.creators.data:
        __create_name__(mods, row, "personal", "creator")
    for row in form.contributors.data:
        __create_name__(mods, row, "personal", "contributor")
    for row in form.corporate_contributors.data:
        __create_name__(mods, row, "corporate", "contributor")
    for row in form.corporate_creators.data:
        __create_name__(mods, row, "corporate", "creator")
    for row in form.languages.data:
        __create_language__(mods, row)
    __create_origin_info__(form, mods) 
    __create_phy_desc__(form, mods)
    __create_subjects__(form, mods) 
    if len(form.title.data) > 0:
        __create_title__(mods, 
            form)
    if pid is not None:
        pid_ident = etree.SubElement(mods,
            "mods:identifer",
            attrib={"type": "local"})
        pid_ident.text = pid
    reparsed = xml.dom.minidom.parseString(
        etree.tostring(mods, "unicode"))
    return reparsed.toprettyxml(encoding='utf-8').decode()

def build_rels_ext(form, pid):
    rels_ext = etree.XML(
        '<rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" />')
    for code, url in DEFAULT_NS.items():
        if code.endswith("rdf"):
            continue
        rels_ext.set("xmlns:{}".format(code), url)
    rdf_description = etree.SubElement(rels_ext,
        "rdf:Description",
        attrib={"rdf:about": "info:fedora/{}".format(pid)})
    if len(form.collection_pid.data) > 0:
        collection_resource = "info:fedora/{}".format(form.collection_pid.data)
        member_of_collection = etree.SubElement(rdf_description,
            "fedora:isMemberOfCollection",
            attrib={"rdf:resource": collection_resource})
    # Add FedoraObject Model
    fedora_model = etree.SubElement(rdf_description,
        "fedora-model:hasModel",
        attrib={"rdf:resource": "info:fedora/fedora-system:FedoraObject-3.0"})
    if len(form.content_models.data) > 0:
        model_str = "info:fedora/{}".format(form.content_models.data)
        content_model = etree.SubElement(rdf_description,
            "fedora-model:hasModel",
            attrib={"rdf:resource": model_str})
    if len(form.parent_pid.data) > 0:
        constituent_str = "info:fedora/{}".format(form.parent_pid.data)
        constituent_of = etree.SubElement(rdf_description,
            "fedora:isConstituentOf",
            attrib={"rdf:resource": constituent_str})
    if len(form.ordering.data) > 0 and len(form.parent_pid.data) > 0:
        sequence_number = etree.SubElement(rdf_description,
            "islandora:isSequenceNumberOf{}".format(form.parent_pid.data.replace(":", "_")))
        sequence_number.text = str(int(form.ordering.data))
    reparsed = xml.dom.minidom.parseString(
        etree.tostring(rels_ext, "unicode"))
    return reparsed.toprettyxml(encoding='utf-8').decode()
        

def create_mods(form):
    """Creates a MODS xml document based on form values

    Args:
        form(flask.request.form): Form
    
    Returns:
        str: Raw XML string of MODS document
    """ 
    mods_context = {'dateCreated': form.get('date_created'),
        'digital_origin': form.get('digital_origin'),
                      'contributors': [],
                      'corporate_contributors': [],
                      'creators': [],
                      'corporate_creators': [],
                      'languages': [],
                      'organizations': [],
                      'schema_type': 'CreativeWork', # Default,
                      'subject_dates': [],
                      'subject_people': [],
                      'subject_places': [],
                      'subject_topics': [],
                      'title': form.get('title'),
                      'typeOfResource': []
                      }
    for row in form.getlist('creators'):
        if len(row) > 0:
            mods_context['creators'].append(row)
    for row in form.getlist('corporate_creators'):
        if len(row) > 0:
            mods_context['corporate_creators'].append(row)
    for row in form.getlist('contributors'):
        if len(row) > 0:
            mods_context['contributors'].append(row)
    for row in form.getlist('corporate_contributors'):
        if len(row) > 0:
            mods_context['corporate_contributors'].append(row)
    for row in form.getlist('subject_people'):
        if len(row) > 0:
            mods_context['subject_people'].append(row)
    for row in form.getlist('organizations'):
        if len(row) > 0:
            mods_context['organizations'].append(row)
    for row in form.getlist('subject_places'):
        if len(row) > 0:
            mods_context['subject_places'].append(row)
    for row in form.getlist('subject_dates'):
        if len(row) > 0:
            mods_context['subject_dates'].append(row)

    for row in form.getlist('subject_topics'):
        if not row is None and len(row) > 0:
            mods_context['subject_topics'].append(row)
    type_res = form.getlist("type_of_resources")
    for row in type_res:
        if len(row) > 0:
            mods_context['typeOfResource'].append(row)
    genre = form.get('genre', '')
    if len(genre) > 0 and not genre.startswith('choose'):
        mods_context['genre'] = genre
    admin_note = form.get('admin_note', '')
    if len(admin_note) > 0:
        mods_context['admin_note'] = admin_note
    description = form.get('description', '')
    if len(description) > 0:
        mods_context['description'] = description
    #alt_title = add_obj_template_form.cleaned_data[
    #      'alt_title']
    #if len(alt_title) > 0:
    #    mods_context['alt_title'] = alt_title
    rights_stmt = form.get('rights_statement', '')
    if len(rights_stmt) > 0:
        mods_context['rights_statement'] = rights_stmt
    mods_context['form'] = form.get('form', '')
    for row in form.getlist("languages"):
        mods_context['languages'].append(row)
    mods_context['publication_place'] = forms.PUBLICATION_PLACE
    mods_context['publisher'] = forms.PUBLISHER
    if len(form.get('extent', '')) > 0:
        mods_context['extent'] = form.get('extent')
    mods_context['subject_topics'] = list(set(mods_context['subject_topics']))
    mods_context['subject_places'] = list(set(
              mods_context['subject_places']))
    mods_xml_template = JINJA_ENV.get_template(
        'templates/fedora_utilities/mods-stub.xml')
    
    mods_xml = mods_xml_template.render(**mods_context)
    return mods_xml

def __new_pid__(fedora_url, auth):
    pid_result = requests.post(
        "{0}new?namespace=coccc".format(fedora_url),
         auth=auth)
    if pid_result.status_code > 399:
        raise ValueError("Could not retrieve nextPID, HTTP Code {}".format(
            pid_result.status_code))
    return pid_result.text

def create_stubs(**kwargs):
    """Function creates 1-n number of basic Fedora Objects in a repository

    Keyword args:
    mods_xml -- MODS XML used for all stub MODS datastreams
    title -- Title of Fedora Object
    parent_pid -- PID of Parent collection
    num_objects -- Number of stub records to create in the parent collection
    content_model -- Content model for the stub records, defaults to
                     compound object
    """
    config = kwargs.get("config")
    mods_xml = kwargs.get('mods_xml')
    title = kwargs.get('title')
    parent_pid = kwargs.get('parent_pid') 
    num_objects = kwargs.get('num_objects')
    content_model = kwargs.get("content_model", 'compoundCModel')
    auth = config.get('FEDORA_AUTH')
    pids = []
    for i in range(0, int(num_objects)):
        new_pid = __new_pid__(config.get("REST_URL"), 
            config.get("FEDORA_AUTH"))
        # Add a label to new PID using the title
        params = urllib.parse.urlencode({"label": title})
        add_label_url = "{0}{1}?{2}".format(
            config.get('REST_URL'),
            new_pid,
            params)
        add_label_result = requests.put(add_label_url,
            auth=auth)
        # Adds MODS datastream to the new object
        params = urllib.parse.urlencode({
            "controlGroup": "M",
            "dsLabel": "MODS",
            "mimeType": "text/xml"})
        new_mods_url = "{0}{1}/datastreams/MODS?{2}".format(
            config.get('REST_URL'),
            new_pid,
            params)
        mods_ds_result = requests.post(new_mods_url,
            data=mods_xml,
            auth=auth)
        # Add RELS-EXT datastream
        rels_ext_template = JINJA_ENV.get_template(
            'templates/fedora_utilities/rels-ext.xml')
        rels_ext_context = {'object_pid':new_pid,
                            'content_model':content_model,
                            'collection_pid':parent_pid}
        rels_ext = rels_ext_template.render(**rels_ext_context)
        params = urllib.parse.urlencode({
            "controlGroup": "M",
            "dsLabel": "RELS-EXT",
            "mimeType": "application/rdf+xml"})
        rels_url = "{0}{1}/datastreams/RELS-EXT?{2}".format(
            config.get('REST_URL'),
            new_pid,
            params)
        rels_result = requests.post(rels_url,
            data=rels_ext,
            auth=auth)
        #yield json.dumps(
        pids.append(
            {"pid": new_pid, 
             "completed": datetime.datetime.utcnow().isoformat()})
    return pids

def generate_stubs(**kwargs):
    return create_stubs(**kwargs)

def new_fedora_object(form, config):
    auth = config.get("FEDORA_AUTH")
    pid = __new_pid__(config.get("REST_URL"), auth)
    params = urllib.parse.urlencode({"label": form.title.data})
    add_label_url = "{0}{1}?{2}".format(
            config.get('REST_URL'),
            pid,
            params)
    add_label_result = requests.put(add_label_url,
         auth=auth)
    # Adds MODS datastream to the new object
    mods_xml = build_mods(form, pid)
    params = urllib.parse.urlencode({
        "controlGroup": "M",
        "dsLabel": "MODS",
        "mimeType": "text/xml"})
    mods_url = "{0}{1}/datastreams/MODS?{2}".format(
         config.get('REST_URL'),
         pid,
         params)
    mods_ds_result = requests.post(mods_url,
         data=mods_xml,
         auth=auth)
    # Adds RELS-EXT 
    rels_ext = build_rels_ext(form, pid)
    params = urllib.parse.urlencode({
        "controlGroup": "M",
        "dsLabel": "RELS-EXT",
        "mimeType": "application/rdf+xml"})
    rels_url = "{0}{1}/datastreams/RELS-EXT?{2}".format(
        config.get('REST_URL'),
        pid,
        params)
    rels_result = requests.post(rels_url,
            data=rels_ext,
            auth=auth)
    # Adds datastream to object if not Compound Content Model type
    if form.digital_object.data is not None: 
        if form.content_models.data.startswith("islandora:compoundCModel"):
            pass
        else:
            datastream = request.files.get(form.digital_object.name)
            label = bibcat.slugify(datastream.name)[0:63]
            new_ds_url = "{}{}/datastreams/OBJ?{}".format(
                config.get("REST_URL"),
                pid,
                urllib.parse.urlencode({"label": label,
                    "controlGroup": "M",
                    "dsLabel": datastream.name,
                    "mimeType": datastream.mimetype}))
            new_ds_result  = requests.post(new_ds_url,
                files={"content": datastream.read()},
                auth=auth)
    else:
        raise ValueError("Error missing Digital Object datastream")
    #! Need to add uploaded thumbnail
    return pid

def repository_move(source_pid,collection_pid):
    """
    Helper view function takes a source_pid and collection_pid,
    retrives source_pid RELS-EXT and updates
    fedora:isMemberOfCollection value with new collection_pid

    :param source_pid: Source Fedora Object PID
    :param collection_pid: Collection Fedora Object PID
    """
    ns = {'rdf':'http://www.w3.org/1999/02/22-rdf-syntax-ns#',
          'fedora':'info:fedora/fedora-system:def/relations-external#'}

    repository = Repository(root=settings.FEDORA_ROOT,
                            username=settings.FEDORA_USER,
                            password=settings.FEDORA_PASSWORD)
    raw_rels_ext = repository.api.getDatastreamDissemination(pid=source_pid,
                                                             dsID='RELS-EXT')
    rels_ext = etree.XML(raw_rels_ext[0])
    collection_of = rels_ext.find('{%s}Description/{%s}isMemberOfCollection' %\
                                  (ns['rdf'],ns['fedora']))
    if collection_of is not None:
        collection_of.attrib['{%s}resource' % ns['rdf']] = "info:fedora/%s" % collection_pid
    repository.api.modifyDatastream(pid=source_pid,
                                    dsID="RELS-EXT",
                                    dsLabel="RELS-EXT",
                                    mimeType="application/rdf+xml",
                                    content=etree.tostring(rels_ext))

def extract_creators(mods_xml):
    """Extracts all creators from a mods_xml file

    Parameters:
    mods_xml -- MODS XML
    """
    creators = []
    names = mods_xml.findall('{{{0}}}name'.format(MODS_NS))
    for name in names:
        creator = None
        role = name.find('{{{0}}}role/{{{0}}}roleTerm'.format(MODS_NS))
        name_parts = name.findall('{{{0}}}namePart'.format(MODS_NS))
        if role is None or role.text != 'creator':
            continue
        for part in name_parts:
            creator = part.text
            creators.append(creator)
    return creators

def extract_title(mods_xml):
    title_entities = []
    titleInfos = mods_xml.findall('{{{0}}}titleInfo'.format(MODS_NS))
    for titleInfo in titleInfos:
        output = {}
        if titleInfo.attrib.get('type')is None:
            # equalvant to MARC 245 $a
            titleValue = titleInfo.find('{{{0}}}title'.format(MODS_NS))
            if titleValue is not None and len(titleValue.text) > 0:
                output['titleValue'] = titleValue.text
                output['label'] = output['titleValue']
            # equalvant to MARC 245 $b
            subtitle = titleInfo.find('{{{0}}}subTitle'.format(MODS_NS))
            if subtitle is not None and len(subtitle.text) > 0:
                output['subtitle'] = subtitle.text
                output['label'] = '{0}: {1}'.format(output.get('label'),
                                                    output['subtitle'])
            # equalivant to MARC 245 $p
            partTitle = titleInfo.find('{{{0}}}partName'.format(MODS_NS))
            if partTitle is not None and len(partTitle.text) > 0:
                output['partTitle'] = partTitle.text
            if len(output) > 0:
                title_entities.append(output)
    return title_entities

def load_edit_form(config, pid):
    mods_url = "{}{}/datastreams/MODS/content".format(
        config.get("REST_URL"),
        pid)
    mods_result = requests.get(mods_url,
        auth=config.get("FEDORA_AUTH"))
    if mods_result.status_code == 404:
        return
    mods_result.encoding = 'utf-8'
    mods_xml = etree.XML(mods_result.text)
    return etree.tostring(mods_xml).decode()

def save_mods_xml(config, pid, mods_xml):
    mods_url = "{}{}/datastreams/MODS?versionable=true".format(
        config.get("REST_URL"),
        pid)
    result = requests.put(mods_url,
        data=mods_xml,
        auth=config.get("FEDORA_AUTH"))
    if result.status_code > 399:
        raise ValueError("Failed to save MODS XML for {}".format(pid))
    return True

# SPARQL Queries
NEWEST_SPARQL = """SELECT DISTINCT ?s ?date
WHERE {{ ?s <fedora-model:createdDate> ?date . }}
ORDER BY DESC(?date)
LIMIT 100
OFFSET {0}"""



"""Module takes Colorado College's MODS XML metadata and extracts information
into a Python dictionary that is converted into JSON and used as an
Elasticsearch document body.

"""
__author__ = "Jeremy Nelson"

import datetime
import rdflib
MODS = rdflib.Namespace("http://www.loc.gov/mods/v3")
MODS_NS = {"mods": str(MODS)}

def calculate_pubyear(rdf_json):
    """Helper function takes RDF json and attempts to extract the 
    publication year based on the date values

    Args:
        rdf_json: RDF JSON dict
    """
    raw_created_date = rdf_json.get("dateCreated")
    try:
        pubdate = datetime.datetime.strptime(raw_created_date, "%Y-%m-%d")
        rdf_json["publicationYear"] = pubdate.year
        return True
    except ValueError:
        if len(raw_created_date) == 4:
            rdf_json["publicationYear"] = raw_created_date
            return True
    except TypeError:
        if raw_created_date is None:
            pass
    if "copyrightYear" in rdf_json and len(rdf_json.get('copyrightYear')) == 4:
        rdf_json["publicationYear"] = rdf_json["copyrightYear"]
        return True
    if raw_created_date and raw_created_date.lower().startswith("unknown"):
        rdf_json["publicationYear"] = raw_created_date.title()
             
    

def generate_field_name(text):
    """Helper function takes text, removes spaces, lowercase for first term,
    and title case for remaining terms before returning a field name for
    indexing into Elasticsearch

    Args:
        text: Text to generate field name
    Returns:
        str: String for field name
    """
    if text is None:
        return
    terms = [s.title() for s in text.split()]
    terms[0] = terms[0].lower()
    return ''.join(terms)


def access_condition2rdf(mods):
    """Function takes MODS XML document and processes mods:accessCondition

    Args:
        mods: MODS etree Document

    Returns:
        dict: Dictionary of extracted Rights statement
    """
    output = {}
    accessCondition = mods.find("mods:accessCondition", MODS_NS)
    if accessCondition is not None:
        if accessCondition.attrib.get("type", "").startswith("useAnd"):
            output["useAndReproduction"] = accessCondition.text
    return output


def language2rdf(mods):
    """Function takes MODS XML document and processes mods:language

    Args:
        mods: MODS etree Document

    Returns:
        dict: Dictionary of extracted language metadata
    """
    output = {'language': []}
    xpath = "mods:language/mods:languageTerm"
    languageTerms = mods.findall(xpath, MODS_NS)
    for row in languageTerms:
        output['language'].append(row.text)
    if len(output['language']) < 1:
        output.pop('language')
    return output

def names2rdf(mods):
    """Function takes MODS XML document and processes mods:names

    Args:
        mods: MODS etree Document

    Returns:
        dict: Dictionary of extracted names
    """
    names = mods.findall("mods:name", MODS_NS)
    output = {}
    for row in names:
        name = row.find("mods:namePart", MODS_NS)
        if hasattr(name, "text") and name.text is None:
            continue
        roleTerm = row.find("mods:role/mods:roleTerm", MODS_NS)
        if roleTerm is not None and hasattr(roleTerm, "text"):
            if roleTerm.text is None:
                continue
            field = generate_field_name(roleTerm.text)
            if field in output:
                output[field].append(name.text)
            else:
                output[field] = [name.text,]
    for key, val in output.items():
        output[key] = list(set(val))
    return output

def notes2rdf(mods):
    """Function takes MODS XML document and processes mods:notes
    
	Args:
        mods: MODS etree Document

    Returns:
        dict: Dictionary of extracted notes
    """
    def process_note(field_name, text):
        """Helper function takes a field name and text checks to see if the
        note is unique in the output dict or adds as the field with the text
        as first member in a list.

        Args:
            field_name: Name of the field to check in output
            text: Text from the note.
        """
        if field_name in output.keys():
            if not text in output[field_name]:
                output[field_name].append(text)
        else:
            output[field_name] = [text,]
    output = {}
    notes = mods.findall("mods:note", MODS_NS)
    for note in notes:
        if not hasattr(note, "text"):
            continue
        note_type = note.attrib.get('type', '')
        if note_type.startswith("admin"):
            process_note("adminNote", note.text)
        if note_type.startswith("thesis"):
            displayLabel = note.attrib.get('displayLabel', '')
            if displayLabel.startswith("Degree"):
                process_note(
                    generate_field_name(displayLabel),
                    note.text)
            else:
                process_note("thesis", note.text)
        else:
            process_note("note", note.text)
    return output


def originInfo2rdf(mods):
    """Function takes MODS XML document and processes mods:originInfo

    Args:
        mods: MODS etree Document

    Returns:
        dict: Dictionary of extracted notes 
    """
    output = {}
    originInfo = mods.find("mods:originInfo", MODS_NS)
    if not originInfo:
        return output
    place = originInfo.find("mods:place/mods:placeTerm", MODS_NS)
    if place is not None and place.text is not None:
        output['place'] = place.text
    publisher = originInfo.find("mods:publisher", MODS_NS)
    if publisher is not None and publisher.text is not None:
        output['publisher'] = publisher.text
    copyrightDate = originInfo.find("mods:copyrightDate", MODS_NS)
    if copyrightDate is not None and copyrightDate.text is not None:
        output["copyrightDate"] = copyrightDate.text
    dateCreated = originInfo.find("mods:dateCreated", MODS_NS)
    if dateCreated is not None and dateCreated.text is not None:
        output["dateCreated"] = dateCreated.text
    dateIssued = originInfo.find("mods:dateIssued", MODS_NS)
    if dateIssued is not None and dateIssued.text is not None:
        output["dateIssued"] = dateIssued.text 
    return output
    

def physicalDescription2rdf(mods):
    """Function takes MODS XML document and processes mods:physicalDescription

    Args:
        mods: MODS etree Document

    Returns:
        dict: Dictionary of extracted extent and digitalOrigin from MODS 
              metadata.
    """
    output = {}
    physicalDescription = mods.find("mods:physicalDescription", MODS_NS)
    if not physicalDescription:
        return output
    extent = physicalDescription.find("mods:extent", MODS_NS)
    if extent is not None and extent.text is not None:
        #! Should add maps and illustrations and page numbers as separate
        #! ES aggregations?
        output['extent'] = extent.text
    digitalOrigin = physicalDescription.find(
        "mods:digitalOrigin", MODS_NS)
    if digitalOrigin is not None and digitalOrigin.text is not None:
        output["digitalOrigin"] = digitalOrigin.text
    return output

def singleton2rdf(mods, element_name):
    """Function takes MODS XML document and processes mods:originInfo

    Args:
        mods: MODS etree Document

    Returns:
        dict: Dictionary of extracted notes 
    """
    output = {}
    output[element_name] = []
    pattern = "mods:{0}".format(element_name)
    elements = mods.findall(pattern, MODS_NS)
    for element in elements:
        if not element.text in output[element_name]:
            output[element_name].append(element.text)
    if len(output[element_name]) > 0:
        return output
    return dict()
    

def subject2rdf(mods):
    """Function takes MODS XML document and processes mods:subject

    Args:
        mods: MODS etree Document

    Returns:
        dict: Dictionary of extracted subjects 
    """
    def process_subject(subject, element_name):
        """Helper function either adds a new subject with the element
	    name, extracts the element from the MODS element and adds a
        unique subject to the output dictionary.
       
        Args:
            subject: MODS subject element
            element_name: element's name
        """
        elements = subject.findall("mods:{0}".format(element_name), MODS_NS)
        for element in elements:
            if hasattr(element, "text"):
                if element_name in output["subject"].keys():
                    if not element.text in output["subject"][element_name]:
                        output["subject"][element_name].append(element.text)
                else:
                    output["subject"][element_name] = [element.text, ]
    output = {"subject":{}}
    subjects = mods.findall("mods:subject", MODS_NS)
    for row in subjects:
        process_subject(row, "genre")
        process_subject(row, "geographic")
        names = row.findall("mods:name", MODS_NS)
        for name in names:
            namePart = name.find("mods:namePart", MODS_NS)
            if namePart and namePart.text is not None:
                if "name" in output['subject'].keys():
                    if not namePart.text in output['subject']['name']:
                        output["subject"]["name"].append(namePart.text)
                else:
                    output["subject"]["name"] = [namePart.text, ]
        process_subject(row, "temporal")
        process_subject(row, "topic")
    return output
        

def title2rdf(mods):
    """
    Function takes a MODS document and returns the titles

	Args:
       mods -- MODS etree XML document
    """
    output = {}
    titles = mods.findall("mods:titleInfo", MODS_NS)
    for row in titles:
        title = row.find("mods:title", MODS_NS)
        type_of = row.attrib.get("type", "")
        if title is None:
            continue
        output["titleRaw"] = title.text
        if type_of.startswith("alt"):
            output["titleAlternative"] = title.text
        else:
            output["titlePrincipal"] = title.text
    return output
   
def url2rdf(mods):
    """
    Function takes a MODS document and returns the urls as 
    the handle.

	Args:
       mods -- MODS etree XML document
    """
    url = mods.find("mods:location/mods:url", MODS_NS)
    #! Saves as handle identifier
    if hasattr(url, "text"):
        return {"handle": url.text}
    return {}

def mods2rdf(mods):
    """Function class all functions to transform a etree MODS doc
    into RDF JSON to be indexed into Elasticsearch.

	Args:
        mods -- MODS etree XML document
    Returns:
        Dictionary of RDF for indexing into Elasticsearch
    """
    rdf_json = {}
    rdf_json.update(singleton2rdf(mods, "abstract"))
    rdf_json.update(access_condition2rdf(mods))
    rdf_json.update(singleton2rdf(mods, "genre"))
    rdf_json.update(language2rdf(mods))
    rdf_json.update(names2rdf(mods))
    rdf_json.update(notes2rdf(mods))
    rdf_json.update(originInfo2rdf(mods))
    rdf_json.update(physicalDescription2rdf(mods))
    rdf_json.update(subject2rdf(mods))
    rdf_json.update(title2rdf(mods))
    rdf_json.update(singleton2rdf(mods, "typeOfResource"))
    rdf_json.update(url2rdf(mods))
    calculate_pubyear(rdf_json)
    return rdf_json



"""indexer contains the Indexer class for indexing Fedora 3.x Objects into
an ElasticSearch instance either by PID or by Collection.

>>> import indexer
>>> test_indexer = indexer.Indexer() # Uses defaults from configuration file
>>> test_indexer.index_pid("abc:1234") # Indexes a single object by PID
>>> test_indexer.index_collection("cba:23") # Indexes all objects in a parent
                                            # collection
"""
__author__ = "Jeremy Nelson"

import datetime
import logging
import os
import requests
import sys
import xml.etree.ElementTree as etree
from copy import deepcopy
from rdflib import Namespace, RDF
from .mods2json import mods2rdf
from .mapping import MAP
from elasticsearch import Elasticsearch

DC = Namespace("http://purl.org/dc/elements/1.1/")
FEDORA_ACCESS = Namespace("http://www.fedora.info/definitions/1/0/access/")
FEDORA = Namespace("info:fedora/fedora-system:def/relations-external#")
FEDORA_MODEL = Namespace("info:fedora/fedora-system:def/model#")
ISLANDORA = Namespace("http://islandora.ca/ontology/relsext#")
etree.register_namespace("fedora", str(FEDORA))
etree.register_namespace("fedora-model", str(FEDORA_MODEL))
etree.register_namespace("islandora", str(ISLANDORA))

logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("Elasticsearch").setLevel(logging.ERROR)


class WebPageHandler(logging.Handler):

    def __init__(self):
        logging.Handler.__init__(self)
        self.messages = []

    def emit(self, record):
        self.messages.append(self.format(record))

    def get_messages(self):
        return self.messages



class Indexer(object):
    """Elasticsearch MODS and PDF full-text indexer for Fedora Repository 3.8"""

    def __init__(self, **kwargs):
        """Initializes an instance of the IndexerClass

		Keyword args:
		    auth -- Tuple of username and password to authenticate to Fedora,
			        defaults to Fedora's standard login credentials

			elasticsearch -- Instance of Elasticsearch Python Client, defaults
			                 to REPO_SEARCH from indexer

            rest_url -- REST URL for Fedora 3.x, defaults to Fedora's El contrabando de El Pasostanard
			ri_url -- SPARQL Endpoint, defaults to Fedora's standard search URL

		"""
        app = kwargs.get("app")
        self.auth = kwargs.get("auth")
        if app and self.auth is None:
            self.auth = app.config.get('FEDORA_AUTH')
        self.elastic = kwargs.get("elasticsearch")
        if app and self.elastic is None:
            self.elastic = app.config.get('ELASTIC_SEARCH')
        if not isinstance(self.elastic, Elasticsearch):
            self.elastic = Elasticsearch(hosts=self.elastic)
        self.islandora_url = kwargs.get("islandora_url")
        if app and self.islandora_url is None:
            self.islandora_url = app.config.get("ISLANDORA_URL")
        self.logger = logging.getLogger(__file__)
        self.messages = []
        self.rest_url = kwargs.get("rest_url")
        if app and self.rest_url is None:
            self.rest_url = app.config.get('REST_URL')
        self.ri_search = kwargs.get("ri_url")
        if app and self.ri_search is None:
            self.ri_search = app.config.get('RI_URL')
        self.skip_pids = []
        # Set defaults if don't exist
        if not self.auth:
            self.auth = ("fedoraAdmin", "fedoraAdmin")
        if not self.rest_url:
            self.rest_url = "http://localhost:8080/fedora/objects/"
        if not self.ri_search:
            self.ri_search = "http://localhost:8080/fedora/risearch"
        if not self.elastic.indices.exists('repository'):
            # Load mapping
            self.elastic.indices.create(index='repository', body=MAP)

    def __add_datastreams__(self, pid):
        """Internal method takes a PID and queries Fedora to extract 
        datastreams and return a list of datastreams names to be indexed
        into Elasticsearch.

		Args:
		   pid -- PID
        """
        ds_pid_url = "{}{}/datastreams?format=xml".format(
            self.rest_url,
            pid)
        result = requests.get(ds_pid_url)
        output = []
        if result.status_code > 399:
            raise IndexerError(
                "Failed to retrieve datastreams for {}".format(pid),
                "Code {} for url {} \nError {}".format(
						result.status_code,
                        ds_pid_url,
                        result.text))
        result_xml = etree.XML(result.text)
        datastreams = result_xml.findall(
            "{{{}}}datastream".format(FEDORA_ACCESS))
        for row in datastreams:
            add_ds = False
            mime_type = row.attrib.get('mimeType')
            if mime_type.startswith("image/tif"):
                output = self.__process_tiff__(pid, datastreams)
                break
            if mime_type.startswith("application/pdf") or\
               mime_type.startswith("audio/mpeg") or\
               mime_type.startswith("audio/mp3") or\
               mime_type.startswith("audio/x-m4a") or\
               mime_type.startswith("video/quicktime") or\
               mime_type.startswith("video/mp4") or\
               mime_type.startswith("image/jpeg") or\
               mime_type.startswith("image/jp2") or\
               mime_type.startswith("audio/wav") or\
               mime_type.startswith("audio/x-wav") or\
               mime_type.startswith("audio/vnd.wave") or\
               mime_type.startswith("application/octet-stream"):
                add_ds = True
            if add_ds:
                output.append({
                    "pid": pid,
                    "label": row.attrib.get('label'),
                    "dsid": row.attrib.get('dsid'),
                    "mimeType": mime_type})
        return output

    def __get_ancestry__(self, pid):
        """Takes a PID and iterates through Fedora for all ancestors
        and returns result as a list.

        Args:
            pid: PID of Fedora Object
        """
        sparql = """SELECT DISTINCT ?s
WHERE {{
  <info:fedora/{0}> <fedora-rels-ext:isMemberOfCollection> ?s .
}}""".format(pid)
        result = requests.post(self.ri_search,
            data={"type": "tuples",
	          "lang": "sparql",
	          "format": "json",
	          "query": sparql},
	    auth=self.auth)
        if result.status_code < 400:
            results = result.json().get('results')
            if len(results) < 1:
                return []
            if len(results) == 1:
                parent_pid = results[0]['s'].split("/")[-1]
                return [parent_pid,] + self.__get_ancestry__(parent_pid)
        else:
            return []


    def __index_compound__(self, pid):
        """Internal method takes a parent PID in a Compound Object and indexes
        all children.

        Args:
            pid -- PID of parent Fedora object
        """
        output = []
        sparql = """SELECT DISTINCT ?s
WHERE {{
   ?s <fedora-rels-ext:isConstituentOf> <info:fedora/{0}> .
}}""".format(pid)
        result = requests.post(
            self.ri_search,
            data={"type": "tuples",
                  "lang": "sparql",
                  "format": "json",
                  "query": sparql},
            auth=self.auth)
        if result.status_code > 399:
            raise IndexerError(
                "Could not retrieve {} constituent PIDS".format(pid),
                "Error code {} for pid {}\n{}".format(
                    result.status_code,
                    pid,
                    result.text))
        for row in result.json().get('results'):
            constituent_pid = row.get('s').split("/")[-1]
            self.skip_pids.append(constituent_pid)
            pid_as_ds = self.__process_constituent__(constituent_pid)
            if pid_as_ds is not None:
                output.extend(pid_as_ds)
        return output

    def __process_constituent__(self, pid, rels_ext=None):
        """Returns constituent PID and returns dictionary compatible with datastream

		Args:
		    pid -- PID
        """
        if not rels_ext:
            rels_ext = self.__get_rels_ext__(pid)
        xpath = "{{{0}}}Description/{{{1}}}isConstituentOf".format(
            RDF,
            FEDORA)
        isConstituentOf = rels_ext.find(xpath)
        parent_pid = isConstituentOf.attrib.get(
            "{{{0}}}resource".format(RDF)).split("/")[-1]
        xpath = "{{{0}}}Description/{{{1}}}isSequenceNumberOf{2}".format(
            RDF,
            ISLANDORA,
			parent_pid.replace(":","_"))
        sequence_number = rels_ext.find(xpath)
        if sequence_number is not None:
            order = sequence_number.text
        else:
            order = "-1"
        datastreams = self.__add_datastreams__(pid)
        for datastream in datastreams:
            datastream['order'] = order
        return datastreams
       
    def  __process_tiff__(self, pid, datastreams):
        """Takes a list of datastreams for an object that contains TIFF
        and attempts to retrieve jpeg derivatives.

        Args:
            pid: PID of Fedora Object with TIFF datastreams
            datastreams: List of datastreams

        Returns: 
            List of datastreams
        """
        output = []
        for row in datastreams:
            mime_type = row.attrib.get('mimeType')
            dsid = row.attrib.get('dsid')
            if mime_type.startswith("image/tif") or \
               dsid.startswith("JPG"):
                output.append(    {"pid": pid,
                    "label": row.attrib.get('label'),
                    "dsid": dsid,
                    "mimeType": mime_type
                    })
        return output
        
        
    def __get_rels_ext__(self, pid):
        """Extracts and returns RELS-EXT base on PID

        Args:
            pid -- PID
        """
        rels_ext_url = "{}{}/datastreams/RELS-EXT/content".format(
            self.rest_url,
            pid)

        rels_ext_result = requests.get(rels_ext_url)
        if rels_ext_result.status_code > 399:
            raise IndexerError("Cannot get RELS-EXT for {}".format(pid),
                "Tried URL {} status code {}\n{}".format(
                    rels_ext_url,
                    rels_ext_result.status_code,
                    rels_ext_result.text))
        return etree.XML(rels_ext_result.text)

    def __get_content_models__(self, pid, rels_ext=None):
        """Extracts and adds content models
        
		Args:
		    pid -- PID
            rels_ext -- XML of RELS-EXT, defaults to None
        """
        if not rels_ext:
            rels_ext = self.__get_rels_ext__(pid)
        output = []
        content_models = rels_ext.findall(
            "{{{0}}}Description/{{{1}}}hasModel".format(
                RDF,
                FEDORA_MODEL))
        if len(content_models) > 0:
            for model in content_models:
                content_model = model.attrib.get("{{{0}}}resource".format(RDF))
                # Remove and save to content_models
                output.append(content_model.split("/")[-1])
        return output
        

    def __reindex_pid__(self, pid, body):
        """Internal method checks and if pid already exists"""
        if self.elastic.count().get('count') < 1:
            return False
        dsl = {
            "query": {
                "term": {"pid": pid}
            }
        }
        result = self.elastic.search(
            body=dsl,
            index='repository',
            doc_type='mods')
        if  result.get('hits').get('total') > 0:
            mods_id = result.get('hits').get("hits")[0].get('_id')
            self.elastic.index(
                id=mods_id,
                index="repository",
                doc_type="mods",
                body=body)
            logging.info("Re-indexed PID=%s, ES-id=%s", pid, mods_id)
            return True


    def incremental_index(self, **kwargs):
        """Performs a incremental index based on date"""
        offset = kwargs.get("offset", 0)
        from_ = kwargs.get("date")
        # Set to 30 days prior if no date
        if from_ is None:
            from_ = datetime.datetime.utcnow() - datetime.timedelta(days=30) 
        latest_100 = requests.post(self.ri_search,
            data={"type": "tuples",
	          "lang": "sparql",
	          "format": "json",
	          "query": NEWEST_SPARQL.format(offset)},
            auth=self.auth)
        search = Search(using=Elasticsearch(hosts=app.config.ELASTIC_SEARCH),
            index="repository")
        total = 0
        for row in latest_100.json().get('results', []):
            pid = row.get('s').split("/")[-1]
            date = datetime.strptime(row.get('date'), "%Y-%m-%dT%H:%M:%S.%fZ")
            if date < from_:
                continue
            search.filter("term", pid=pid)
            found_pid = search.execute()
            if len(found_pid) > 0:
                # Try next shard
                offset +=1
                total += self.incremental_index(offset=offset)
            else:
                ancestors = self.__get_ancestry__(pid)
                self.index_pid(pid, parent=ancestors[0], inCollections=ancestors)
                total += 1
        return total

    def index_pid(self, pid, parent=None, inCollections=[]):
        """Method retrieves MODS and any PDF datastreams and indexes
        into repository's Elasticsearch instance

        Args:
            pid: PID to index
            parent: PID of parent collection, default is None
            inCollections: List of pids that this object belongs int, used for
			    aggregations.

        Returns:
            boolean: True if indexed, False otherwise
        """
        rels_ext = self.__get_rels_ext__(pid)
        xpath = "{{{0}}}Description/{{{1}}}isConstituentOf".format(
            RDF,
            FEDORA)
        is_constituent = rels_ext.find(xpath)
        # Skip and don't index if pid is a constituent of another compound 
		# object
        if is_constituent is not None:
            return False
        # Quick check to see Islandora is available, skip if error code
        # is 403 Access Denied
        islandora_url = "{}{}".format(self.islandora_url,
                                      pid)
        islandora_result = requests.get(islandora_url)
        if islandora_result.status_code == 403:
            logging.info("Access denied {}".format(pid))
            return False
        # Extract MODS XML Datastream
        mods_url = "{}{}/datastreams/MODS/content".format(
            self.rest_url,
            pid)
        mods_result = requests.get(
            mods_url,
            auth=self.auth)
        mods_result.encoding = 'utf-8'
        if mods_result.status_code > 399:
            err_title = "Failed to index PID {}, error={} url={}".format(
                pid,
                mods_result.status_code,
                mods_url)
            logging.error(err_title)
            # 404 error assume that MODS datastream doesn't exist for this
			# pid, return False instead of raising IndexerError exception
            if mods_result.status_code == 404:
                return False
            raise IndexerError(
                err_title,
                mods_result.text)
        try:
            if not isinstance(mods_result.text, str):
                mods_xml = etree.XML(mods_result.text.decode())
            else:
                mods_xml = etree.XML(mods_result.text)
        except etree.ParseError:
            msg = "Could not parse pid {}".format(pid)
            return False
        mods_body = mods2rdf(mods_xml)
        # Extract and process based on content model
        mods_body["content_models"] = self.__get_content_models__(pid, rels_ext)
        mods_body['pid'] = pid
        # Used for scoping aggregations
        if len(inCollections) > 0:
            mods_body['inCollections'] = inCollections
        # Used for browsing
        if parent:
            mods_body['parent'] = parent
        # Delete any existing document for the pid
        delete_existing_dsl = {
            "query": {
                "term": {"pid": pid}
            }
        }
        self.elastic.delete_by_query(body=delete_existing_dsl,
            index='repository')
        # Add Datasteams to Index
        # Extract Islandora Content Models from REL-EXT 
        if "islandora:compoundCModel" in mods_body["content_models"]:
            mods_body["datastreams"] = self.__index_compound__(pid)
        else: 
            mods_body["datastreams"] = self.__add_datastreams__(pid)
        mods_index_result = self.elastic.index(
            index="repository",
            doc_type="mods",
            body=mods_body)
        mods_id = mods_index_result
        if mods_id is not None:
            msg = "Indexed PID={0}, ES-id={1}".format(
                pid,
                mods_id.get('_id'))
            logging.info(msg)
            return True
        return False

    def index_collection(self, pid, parents=[]):
        """Method takes a parent collection PID, retrieves all children, and
        iterates through and indexes all pids

        Args:
            pid -- Collection PID
            parents -- List of all Fedora Object PIDs that pid is in the 
	               collection

        """
        sparql = """SELECT DISTINCT ?s
WHERE {{
  ?s <fedora-rels-ext:isMemberOfCollection> <info:fedora/{}> .
}}""".format(pid)
        started = datetime.datetime.utcnow()
        msg = "Started indexing collection {} at {}".format(
            pid,
            started.isoformat())
        self.logger.info(msg)
        self.messages.append(msg)
        children_response = requests.post(
            self.ri_search,
            data={"type": "tuples",
                  "lang": "sparql",
                  "format": "json",
                  "query": sparql},
            auth=self.auth)
        if children_response.status_code < 400:
            children = children_response.json().get('results')
            for row in children:
                iri = row.get('s')
                child_pid = iri.split("/")[-1]
                child_parents = deepcopy(parents)
                child_parents.append(pid)
                self.index_pid(child_pid, pid, child_parents)
                is_collection_sparql = """SELECT DISTINCT ?o
WHERE {{        
  <info:fedora/{0}> <fedora-model:hasModel> <info:fedora/islandora:collectionCModel> .
  <info:fedora/{0}> <fedora-model:hasModel> ?o
}}""".format(child_pid)
                is_collection_result = requests.post(
                    self.ri_search,
                    data={"type": "tuples",
                          "lang": "sparql",
                          "format": "json",
                          "query": is_collection_sparql},
                    auth=self.auth)
                if len(is_collection_result.json().get('results')) > 0:
                    self.index_collection(child_pid, child_parents)
        else:
            err_title = "Failed to index collection PID {}, error {}".format(
                pid,
                children_response.status_code)
            logging.error(err_title)
            raise IndexerError(
                err_title,
                children_response.text)
        end = datetime.datetime.utcnow()
        msg = "Indexing done {} at {}, total object {} total time {}".format(
            pid,
            end.isoformat(),
            len(children),
            (end-started).seconds / 60.0)
        self.logger.info(msg)
        self.messages.append(msg)

    def reset(self):
        """Deletes existing repository index and reloads Map"""
        if self.elastic.indices.exists('repository'):
            self.elastic.indices.delete(index='repository')
            # Load mapping
            self.elastic.indices.create(index='repository', body=MAP)


class IndexerError(Exception):
    """Base for any errors indexing Fedora 3.x objects into Elasticsearch"""

    def __init__(self, title, description):
        """Initializes an instance of IndexerError

	    Args:
	       title -- Title for Error
		   description -- More detailed information about the exception
        """
        super(IndexerError, self).__init__()
        self.title = title
        self.description = description

    def __str__(self):
        """Returns string representation of the object using the instance's
		title"""
        return repr(self.title)



"""
 mod:`urls` Fedora Batch App URL rourting
"""
__author__ = "Jeremy Nelson"

import json
import time
import threading

from concurrent.futures import ThreadPoolExecutor
from elasticsearch import Elasticsearch
from flask import Flask, render_template, request, redirect, Response
from flask import jsonify, flash, url_for, abort, session
from .forms import AddFedoraBatchFromTemplate, AddFedoraObjectFromTemplate
from .forms import MODSReplacementForm, MODSSearchForm, IndexRepositoryForm
from .forms import EditFedoraObjectFromTemplate, LoadMODSForm 
from .helpers import create_mods, generate_stubs, load_edit_form, build_mods
from .helpers import save_mods_xml, build_rels_ext, new_fedora_object
from .indexer import Indexer
from .repairer import update_multiple, update_mods

app = Flask(__name__, instance_relative_config=True)
app.config.from_pyfile('config.py')


#executor = ThreadPoolExecutor(2)

ACTIVE_MSGS = []
BACKEND_THREAD = None
MODS_NS = {'mods': 'http://www.loc.gov/mods/v3'}

class IndexerThread(threading.Thread):

    def __init__(self, **kwargs):
        threading.Thread.__init__(self)
        self.indexer = Indexer(app=app, 
            elasticsearch=kwargs.get('elasticsearch'))
        self.pid = kwargs.get("pid")
        self.job = kwargs.get("job")
        
    def run(self):
        if self.job.lower().startswith("full"):
            self.indexer.reset()
            self.indexer.index_collection(self.pid)
        return 

class IndexerServerSideEvent(object):

    def __init__(self, **kwargs):
        self.job = kwargs.get("job")
        self.pid = kwargs.get("pid")
        self.data = kwargs.get("message")
        self.desc_map = {
            self.job: "job",
            self.pid: "pid",
            self.data: "data"
        }
       

    def encode(self):
        if not self.data:
            return ""
        lines = []
        for key, val in self.desc_map.items():
            if key:
                lines.append("{0}: {1}".format(val, key))
        return "{}\n\n".format("\n".join(lines))

class RepairerThread(threading.Thread):

    def __init__(self, **kwargs):
        threading.Thread.__init__(self)
        self.pid_listing = kwargs.get('pid_listing')
        self.xpath = kwargs.get('xpath')
        self.old_value = kwargs.get('old_value')
        self.new_value = kwargs.get('new_value')
        
    def run(self):
        update_multiple(pid_list=self.pid_listing, 
            xpath=self.xpath,
            old_value=self.old_value,
            new_value=self.new_value)
        return

@app.route("/")
def default():
    return render_template('fedora_utilities/app.html')

@app.route("/about")
def about():
    return render_template("fedora_utilities/about.html")

@app.route("/add_object", methods=["GET", "POST"])
def add_object():
    object_form = AddFedoraObjectFromTemplate(csrf_enabled=False)
    if object_form.validate_on_submit():
        pid = new_fedora_object(object_form, app.config)
        time.sleep(10) # Needed for new object to propagate to Fedora
        # Directly indexing into Digital CC Elasticsearch
        __index_pid__(app.config.get("ELASTIC_SEARCH"), pid)
        msg = """Added PID: {0} available at <a href="{1}pid/{0}">{1}pid/{0}</a>""".format(
            pid,
            app.config.get("DIGITAL_CC_URL"))
        flash(msg)
        return redirect(url_for('add_object'))
    else:
        return render_template("fedora_utilities/object-ingest.html",
            object_form=object_form)
        

@app.route("/add_stub", methods=["GET", "POST"])
def add_stub():
    ingest_form = AddFedoraBatchFromTemplate(csrf_enabled=False)
    if ingest_form.validate_on_submit():
        mods_xml = create_mods(request.form)
        return jsonify(generate_stubs(
            config=app.config,
            mods_xml=mods_xml,
            title=request.form.get('title'),
            parent_pid=request.form.get('collection_pid'),
            num_objects=request.form.get('number_objects'),
            content_model=request.form.get('content_models')))
    return render_template('fedora_utilities/batch-ingest.html',
        ingest_form=ingest_form)

@app.route("/edit/", methods=["GET", "POST"])
@app.route("/edit/<path:pid>", methods=["GET", "POST"])
def edit_mods(pid=None):
    load_form, mods_xml = None, None
    if pid is None:
        load_form = LoadMODSForm()
        if load_form.validate_on_submit():
            return redirect(url_for("edit_mods", pid=load_form.pid.data))
    else:
        mods_xml = load_edit_form(app.config, pid)
        if mods_xml is None:
            return abort(404)
        if request.method.startswith("POST"):
            mods_xml = build_mods(edit_form)
            return mods_xml
    return render_template('fedora_utilities/mods-edit.html',
        load_form=load_form,
        pid=pid,
        mods_xml=mods_xml)

@app.route("/edit/<pid>/save", methods=["POST"])
def save_mods(pid):
    mods_xml = request.form.get("xml")
    mods_xml = mods_xml.replace('xml:space="preserve"', "")
    save_mods_xml(app.config, pid, mods_xml)
    return jsonify({"message": "MODS XML saved",
                    "pid": pid})

@app.route("/index/status")
def indexing_status():
    if BACKEND_THREAD is None:
        # Check in 15 seconds
        #time.sleep(15)
        #if BACKEND_THREAD is None:
        #    msg     
        return jsonify({"message": "Indexer doesn't exist"})
    # Arbitrary limit of 50 continue
    elif session['continue'] > app.config.get('CONTINUE_LIMIT', 50):
       return jsonify({"message": "Indexing complete"}) 
    else:
        if len(BACKEND_THREAD.indexer.messages) > 0:
            msg = BACKEND_THREAD.indexer.messages.pop(0)
            if session['continue'] > 0:
                session['continue'] -= 1
        else:
            msg = "&hellip;continue indexing {0}".format(
                BACKEND_THREAD.pid)
            session['continue'] += 1
        ev = IndexerServerSideEvent(message=msg, 
            job=BACKEND_THREAD.job,
            pid=BACKEND_THREAD.pid)
        return Response(ev.encode(),
                        mimetype="text/event-stream")
    

def __index_pid__(search_idx, pid):
    indexer = Indexer(app=app, 
        elasticsearch=search_idx)
    ancestors = indexer.__get_ancestry__(pid)
    ancestors.reverse()
    indexer.index_pid(pid, ancestors[-1], ancestors)


@app.route("/index/pid", methods=["POST"])
def index_pid():
    pid = request.form.get("pid")
    search_index = request.form.get("indices")
    __index_pid__(search_index, pid)
    flash("Indexed PID {} in index {}".format(pid, search_index))
    return redirect(url_for('index_repository'))

@app.route("/index", methods=["POST", "GET"])
def index_repository():
    global BACKEND_THREAD
    index_form = IndexRepositoryForm(csrf_enabled=False)
    session['continue'] = 0
    if index_form.validate_on_submit():
        if index_form.index_choice.data.startswith("0"):
            return jsonify({"message": "Started Incremental Indexing"})
        elif index_form.index_choice.data.startswith("1"):
            elastic_host = index_form.indices.data
            BACKEND_THREAD = IndexerThread(
                elasticsearch=elastic_host,
                job="full", 
                pid=app.config.get("INITIAL_PID"))
            BACKEND_THREAD.start()
            return jsonify({"message": "Started Full Indexing"})
        else:
            return jsonify({"message": "Unknown Indexing option"})
    return render_template('fedora_utilities/index-repository.html',
        index_form=index_form)


@app.route("/mods-replacement", methods=["POST", "GET"])
def mods_replacement():
    replace_form = MODSReplacementForm(csrf_enabled=False)
    search_form = MODSSearchForm(csrf_enabled=False)
    if replace_form.validate_on_submit():
        collection_pid = replace_form.collection_pid.data
        pid_listing = replace_form.pid_listing.data.split(",")
        search_index = replace_form.indices.data
        xpath = replace_form.select_xpath.data
        old_value = replace_form.old_value.data
        new_value = replace_form.new_value.data
        if collection_pid is not None and len(collection_pid) > 0:
            pid_listing = get_collection_pids(app.config, collection_pid)
        for pid in pid_listing:
            update_mods(app=app,
                pid=pid,
                xpath=xpath,
                old=old_value,
                new=new_value)
        # Kludge to allow changes to propagate through Fedora
        time.sleep(10)
        print("Finished updating all Fedora Objects, trying to index now")
        for pid in pid_listing:
            __index_pid__(search_index, pid)
        flash("Replaces {} old {} with {} for pid {}".format(
            xpath,
            old_value,
            new_value,
            pid_listing))
                   #return jsonify({"total": len(pid_listing)})
        return redirect(url_for('mods_replacement'))
    return render_template('fedora_utilities/mods-replacement.html',
        replace_form=replace_form,
        search_form=search_form)

@app.route("/search", methods=["POST", "GET"])
def search_pids():
    search_form = MODSSearchForm(csrf_enabled=False)
    if search_form.validate_on_submit():
        es_host = search_form.indices.data
        es = Elasticsearch(hosts=[es_host])
        found_pids = []
        query = search_form.query.data
        facet = search_form.facet.data
        dsl = {"_source": ["pid"]}
        if facet.startswith("none"):
            dsl["query"] = {
                "match": {
                    "_all": query
                }    
            }
        else:
            dsl = {
                "query": {
                    "term": { facet: query }
                }
            }
        search_results = es.search(body=dsl, 
                index='repository',
                size=50)
        for hit in search_results.get('hits', {})\
            .get('hits', []):
            pid = hit.get('_source', {}).get('pid')
            found_pids.append(pid)
        return jsonify({
            "pids": found_pids, 
            "total": search_results.get('hits', {}).get('total', 0)})
    return jsonify({"pids": []})
    

if __name__ == '__main__':
    app.run(host='0.0.0.0', debug=True, port=9455)



__author__ = "Jeremy Nelson"
import datetime
from flask_wtf import FlaskForm as Form
from wtforms import SelectField, StringField, TextAreaField
from wtforms.fields.html5 import DateField
from wtforms.fields import FieldList
from wtforms import validators, FileField

DIGITAL_ORIGIN = [('born digital','born digital'),
                  ('reformatted digital', 'reformatted digital'),
                  ('digitized microfilm', 'digitized microfilm'),
                  ('digitized other analog', 'digitized other analog')]

FACETS = [("none", "None"),
          ("typeOfResource", "Format"),
          ("subject.geographic", "Geographic"),
          ("genres", "Genres"),
          ("language", "Languages"),
          ("publicationYear", "Publication Year"),
          ("subject.temporal", "Temporal (Time)"),
          ("subject.topic", "Topic")] 

GENRE = [('choose', 'Choose...'),
         ("abstract or summary", "Abstract or Summary"),
         ("art original", "Art Original"),
         ("art reproduction", "Art Reproduction"),
         ("article", "Article"),
         ("atlas", "Atlas"),
         ("autobiography ", "Autobiography "),
         ("bibliography", "Bibliography"),
         ("biography", "Biography"),
         ("book", "Book "),
         ("calendar", "Calendar"),
         ("catalog", "Catalog"),
         ("chart", "Chart"),
         ("comic or graphic novel ", "Comic or Graphic Novel "),
         ("comic strip", "Comic Strip"),
         ("conference publication", "Conference Publication"),
         ("database", "Database"),
         ("dictionary", "Dictionary"),
         ("diorama", "Diorama"),
         ("directory", "Directory"),
         ("discography", "Discography"),
         ("drama", "Drama"),
         ("encyclopedia", "Encyclopedia"),
         ("essay", "Essay"),
         ("festschrift", "Festschrift"),
         ("fiction", "Fiction"),
         ("filmography", "Filmography"),
         ("filmstrip", "Filmstrip"),
         ("finding aid ", "Finding Aid "),
         ("flash card", "Flash Card"),
         ("folktale ", "Folktale "),
         ("font", "Font"),
         ("game", "Game"),
         ("government publication ", "Government Publication "),
         ("graphic", "Graphic"),
         ("globe", "Globe"),
         ("handbook", "Handbook"),
         ("history ", "History "),
         ("hymnal", "Hymnal"),
         ("humor, satire", "Humor, Satire"),
         ("index", "Index"),
         ("instruction ", "Instruction "),
         ("interview ", "Interview "),
         ("issue", "Issue"),
         ("journal", "Journal"),
         ("kit", "Kit"),
         ("language instruction", "Language Instruction"),
         ("law report or digest", "Law Report or Digest"),
         ("legal article", "Legal Article"),
         ("legal case and case notes", "Legal Case and Case Notes"),
         ("legislation", "Legislation"),
         ("letter ", "Letter "),
         ("loose-leaf ", "Loose-Leaf "),
         ("map", "Map"),
         ("memoir ", "Memoir "),
         ("microscope slide", "Microscope Slide"),
         ("model", "Model"),
         ("motion picture", "Motion Picture"),
         ("multivolume monograph", "Multivolume Monograph"),
         ("newspaper", "Newspaper"),
         ("novel ", "Novel "),
         ("numeric data", "Numeric Data"),
         ("offprint", "Offprint"),
         ("online system or service", "Online System or Service"),
         ("patent", "Patent"),
         ("periodical", "Periodical"),
         ("picture", "Picture"),
         ("poetry ", "Poetry "),
         ("programmed text", "Programmed Text"),
         ("realia", "Realia"),
         ("rehearsal ", "Rehearsal "),
         ("remote sensing image", "Remote Sensing Image"),
         ("reporting ", "Reporting "),
         ("review", "Review"),
         ("script", "Script"),
         ("series", "Series"),
         ("short story", "Short Story"),
         ("slide", "Slide"),
         ("sound ", "Sound "),
         ("speech", "Speech"),
         ("standard or specification", "Standard or Specification"),
         ("statistics", "Statistics"),
         ("survey of literature", "Survey of Literature"),
         ("technical drawing", "Technical Drawing"),
         ("technical report", "Technical Report"),
         ("thesis", "Thesis"),
         ("toy", "Toy"),
         ("transparency", "Transparency"),
         ("treaty", "Treaty"),
         ("videorecording ", "Videorecording "),
         ("web site", "Web Site"),
         ("yearbook", "Yearbook")]
         

CONTENT_MODELS = [('islandora:sp_basic_image', 'Basic Image Content Model'),
                  ('islandora:sp_pdf', 'PDF Content Model'),
                  ('islandora:compoundCModel', 'Compound Object Content Model'),
                  ('islandora:sp-audioCModel', 'Audio Content Model'),
                  ('islandora:sp_videoCModel', 'Video Content Model')]

INSTITUTION_NAME = 'Colorado College'

LANGUAGES = [('English','English'),
             ('Spanish', 'Spanish'),
             ('French', 'French'),
             ('German', 'German'),
             ('Italian', 'Italian'),
             ('Chinese', 'Chinese'),
             ('Japanese', 'Japanese')] 

MARC_FREQUENCY = [('choose', 'Choose...'),
                  ('Semiweekly', 'Semiweekly - 2 times a week'),
                  ('Three times a week', 'Three times a week'),
                  ('Weekly', 'Weekly'),
                  ('Biweekly', 'Biweekly - every 2 weeks'),
                  ('Three times a month', 'Three times a month'),
                  ('Semimonthly', 'Semimonthly - 2 times a month'),
                  ('Monthly', 'Monthly'),
                  ('Bimonthly', 'Bimonthly - every 2 months'),
                  ('Quarterly', 'Quarterly'),
                  ('Three times a year', 'Three times a year'),
                  ('Semiannual', 'Semiannual - 2 times a year'),
                  ('Annual', 'Annual'),
                  ('Biennial', 'Biennial - every 2 years'),
                  ('Triennial', 'Triennial - every 3 years'),
                  ('Completely irregular', 'Completely irregular')]

TYPE_OF_RESOURCE = [('text', 'Text or Language Material'),
                    ('cartographic', 'Maps or other Cartographic Materials'),
                    ('notated music', 'Notated Music'),
                    ('sound recording', 'Sound Recording, mixture of music and non-music recording'),
                    ('sound recording-musical', 'Musical Sound Recording'),
                    ('sound recording-nonmusical', 'Spoken word or other Non-music sound recording'),
                    ('still image', 'Still Image'),
                    ('moving image', 'Video or Film, Moving Image'),
                    ('three dimensional object', 'Three Dimensional Object'),
                    ('software, multimedia', 'Software or other electronic Multimedia'),
                    ('mixed material', 'Mixed Material')]

RIGHTS_STATEMENT = "Copyright restrictions apply. Contact Colorado College for permission to publish."
PLACE = 'Colorado Springs (Colo.)'
PUBLISHER = "Colorado College"
PUBLICATION_PLACE = 'Colorado Springs, Colorado'
SEARCH_INDICES = [('li-testdocker1', 'Test Environment (li-testdocker1)'),
                  ('li-docker1', 'Production Environment (li-docker1)')]

class FedoraObjectFromTemplate(Form):
    abstract = TextAreaField("Abstract",
        description="""A summary of the content of the resource.""",
        validators=[validators.length(max=1500)])
    admin_notes = FieldList(
        TextAreaField('Administrative Notes',
            validators=[validators.length(max=1500)]),
        description="""Administrative Note related to the collection, cataloging,
or management of the resource.""",
        min_entries=1)
    alt_title = StringField('Alternative Title',
        description="""Varying form of the title if it contributes to the 
further identification of the item""")
    collection_pid = StringField("PID of Collection",
        description="""PID of the Collection that the resource belongs too.""",
        validators=[validators.required(), validators.length(max=20)])
    content_models = SelectField('Islandora Content Model',
        choices=CONTENT_MODELS,
        default='compoundCModel',
        validators=[validators.required()])
    contributors = FieldList(StringField("Contributors"), 
        description="""A person responsible for making contributions to the 
resource.""",
        min_entries=1)
    corporate_contributors = FieldList(StringField("Corporate Contributors"), 
        description="""An organization responsible for making contributions to the 
resource.""",
        min_entries=1)
    corporate_creators = FieldList(StringField("Corporate Creators"), 
        description="""An organization responsible for the intellectual or 
artistic content of a resource.""", 
        min_entries=1)
    creators = FieldList(
        StringField("Creators"), 
        min_entries=1,
        description="""A person responsible for the intellectual or artistic 
content of a resource.""" 
    )
    date_created = StringField('Date Created', 
        description="""Date of object creation, NOT the date metadata was entered. 
Use YYYY-MM-DD, YYYY-MM, or YYYY formats""")
    digital_origin = SelectField('Digital Origin',
        choices=DIGITAL_ORIGIN,
        description="""The method by which a resource achieved digital form.""")
    description = TextAreaField('Description',
        description="""General information relating to the physical description of 
the resource.""",
        validators=[validators.optional(), 
            validators.length(max=1500)])
    extent = TextAreaField('Extent',
        description="""A statement of the number and specific material of the units 
of the resource that express physical extent.""",
        validators=[validators.optional(), 
                    validators.length(max=1500)])
    form = StringField('Form')
    frequency_free_form = StringField('Other')
    frequency = SelectField('Frequency',
        choices=MARC_FREQUENCY)
    genre = SelectField('Genre',
        choices=GENRE,
        description="""A term or terms that designate a category characterizing
a particular style, form, or content, such as artistic, musical, literary 
composition, etc.""")
    languages = FieldList(SelectField('Languages', choices=LANGUAGES), 
        min_entries=1,
        description="""A designation of the language in which the content of 
a resource is expressed, NOT the language of cataloging.""")
    ordering = StringField("Order in Compound Object",
        description="Child display order within a Compound Object")
    organizations = StringField("Organizations",
                                validators=[validators.optional(), 
                                            validators.length(max=255)],
                                default=INSTITUTION_NAME)
    parent_pid = StringField("PID of Parent",
        description="""PID of parent object, parent object 
must have the Compound Content Model""")
    publisher = StringField('Publisher', default=PUBLISHER,
        description="""The name of the entity that published, printed, 
distributed, released, issued, or produced the resource""")
    publication_place = StringField("Place of Publication", 
        default=PUBLICATION_PLACE,
        description="""Name of a place associated with the issuing, publication, 
release, distribution, manufacture, production, or origin of a resource""")
    rights_statement = TextAreaField('Rights Statement',
                                      default=RIGHTS_STATEMENT)
    subject_dates = FieldList(StringField('Subject -- Dates'),
        description="""Used for chronological subject terms or temporal coverage.""",
        min_entries=1)

    subject_orgs = FieldList(StringField('Subject -- Organization'),
        description="""A name of an organization that is used as a subject.""",
        label='Subject Organizations',
        min_entries=1)
    subject_people = FieldList(StringField('Subject -- Person'),
        description="""A name of a Person used as a subject.""",
        min_entries=1) 
    subject_places = FieldList(
        StringField('Subject -- Places',
            default=PLACE),
        min_entries=1, 
        description="""Used for geographic subject terms that are not parsed as hierarchical 
    geographics"""
    )
    subject_topics = FieldList(StringField('Subject -- Topic'),
        description=""""topic" is used as the tag for any topical subjects that are not appropriate in the 
<geographic>, <temporal>, <titleInfo>, or <name> subelements.""",
        min_entries=1)
    title = StringField('Title',
        description="""A word, phrase, character, or group of characters that 
constitutes the chief title of a resource, i.e., the title normally used when 
citing the resource.""",
            validators=[validators.length(max=120), validators.required()])
    type_of_resources = FieldList(SelectField('Type of Resource', 
                                              choices=TYPE_OF_RESOURCE),
        min_entries=1,
        description="""A term that specifies the characteristics and general type 
of content of the resource.""")

class AddFedoraObjectFromTemplate(FedoraObjectFromTemplate):
    digital_object = FileField(description="Primary datastream for resource")
    thumbnail = FileField(description="Optional thumbnail image for resource")

class AddFedoraBatchFromTemplate(FedoraObjectFromTemplate):
    number_objects = StringField('Number of stub records', default=1)

class EditFedoraObjectFromTemplate(FedoraObjectFromTemplate):
    pid = StringField("PID")

class IndexRepositoryForm(Form):
    index_choice = SelectField('Incremental or Full',
        choices=[('0', 'Incremental'),( '1', 'Full')])
    indices = SelectField('Search Index',
        choices=SEARCH_INDICES)
    start_from = DateField(
        'Start Index From', 
        format='%Y-%m-%d',
        validators=(validators.Optional(),))

class LoadMODSForm(Form):
    pid = StringField("Enter PID to load MODS",
        validators=[validators.optional(),])


class MODSReplacementForm(Form):
    old_value = StringField("Old Value")
    new_value = StringField("New Value")
    collection_pid = StringField("Collection PID",
        validators=[validators.optional(),])
    indices = SelectField('Search Index',
        choices=SEARCH_INDICES)
    pid_listing = TextAreaField('PID Listing',
            validators=[validators.optional(),])
    select_xpath = StringField("Selection XPath")

class MODSSearchForm(Form):
    facet = SelectField('Facet',
        choices=FACETS)
    indices = SelectField('Search Index',
        choices=SEARCH_INDICES)
    query = StringField("Search")





"""Module repairs errors with MODS metadata"""
__author__ = "Jeremy Nelson, Sarah Bogard"

import datetime
import os
import requests
import sys
import zipfile

import xml.etree.ElementTree as etree
import rdflib
import urllib.parse

MODS = rdflib.Namespace("http://www.loc.gov/mods/v3")
etree.register_namespace("mods", str(MODS))

BASE_DIR = os.path.dirname(os.path.dirname(__file__))
sys.path.append(BASE_DIR)

class RepairMODSError(Exception):

    def __init__(self, pid, message):
        self.pid = pid
        self.message = message

    def __str__(self):
        return "Error with {}'s MODS\n{}".format(self.pid, self.message)

def add_mods(**kwargs):
    """Takes a pid and a populated MODS element"""

def update_mods(**kwargs):
    """Function takes pid, field, and old_value, replaces with
	a new_value.

    Args:
        pid -- 
		field_xpath -- 
		old_value -- 
		new_value --
    """
    app = kwargs.get('app')
    pid = kwargs.get("pid")
    field_xpath = kwargs.get("xpath") 
    old_value = kwargs.get("old")
    new_value = kwargs.get("new")
    start = datetime.datetime.utcnow()
    rest_url = kwargs.get("rest_url", app.config.get("REST_URL"))
    auth = kwargs.get("auth", app.config.get("FEDORA_AUTH"))
    mods_base_url = "{}{}/datastreams/MODS".format(
        rest_url,
        pid.strip())
    get_mods_url = "{}/content".format(mods_base_url)
    mods_result = requests.get(
        get_mods_url,
        auth=auth)
    if mods_result.status_code > 399:
        err_title = """"Failed to replace {} with {} for PID {}, 
error={} url={}""".format(
            old_value, 
            new_value, 
            pid,
            mods_result.status_code,
            get_mods_url)
    raw_xml = mods_result.text
    mods_xml = etree.XML(raw_xml)
    old_value_elements = mods_xml.findall(field_xpath, {"mods": str(MODS)})
    for element in old_value_elements:
        if element.text == old_value:
            element.text = new_value
    url_params = {"controlGroup": "M",
        "dsLabel": "MODS",
        "mimeType": "text/xml",
        "versionable": "true"}
    mods_update_url = "{}?{}".format(
        mods_base_url,
        urllib.parse.urlencode(url_params))
    raw_xml = etree.tostring(mods_xml)
    put_result = requests.post(
        mods_update_url,
	files={"content":  raw_xml},
        auth=app.config.get('FEDORA_AUTH'))
    if put_result.status_code < 300:
        return True
    else:
        raise RepairMODSError(
            pid, 
            "Failed to update MODS with PUT\nStatus code {}\n{}".format(
                put_result.status_code,
                put_result.text))

def update_multiple(**kwargs):
    """Function takes a list of PIDs, the MODS field xpath, the old value to be 
    replaced by the new value.

    Keyword args:
        pid_list -- Listing of PIDs
        field_xpath -- Field XPath
        old_value -- Old string value to be replaced
        new_value -- New string value
    """
    app = kwargs.get('app')
    pid_list = kwargs.get('pid_list')
    field_xpath = kwargs.get('xpath')
    old_value = kwargs.get('old_value')
    new_value = kwargs.get('new_value') 
    start = datetime.datetime.utcnow()
    msg = "Starting MODS update for {} PIDS at {}".format(
        len(pid_list), 
        start.isoformat())
    yield msg
    errors = []
    for i, pid in enumerate(pid_list):
        if not update_mods(pid, field_xpath, old_value, new_value):
            errors.append(pid)
        if not i%25:
            #print(i, end="")
            yield i
        elif not i%10:
            #print(".", end="")
            yield "."
    end = datetime.datetime.utcnow()
    msg = "Finished updating MODS for {}, errors {} at {}, total {}".format(
        len(pid_list),
        len(errors),
        end.isoformat(),
        (end-start).seconds / 60.0)
    yield msg


def get_collection_pids(config, pid):
    """Function takes a Flask app config and a collection PID,
    returns all children PIDS

    Args:
        config: Flask app config
        pid: PID of collection
    """
    output = []
    sparql = """SELECT DISTINCT ?s
WHERE {{
  ?s <fedora-rels-ext:isMemberOfCollection> <info:fedora/{}> .
}}""".format(pid)
    children_response = requests.post(
            config.get("RI_SEARCH"),
            data={"type": "tuples",
                  "lang": "sparql",
                  "format": "json",
                  "query": sparql},
            auth=config.get("FEDORA_AUTH"))
    children = children_response.json().get('results')
    for row in children:
        iri = row.get('s')
        child_pid = iri.split("/")[-1]
        output.append(child_pid)
    return output



"""Setup module for Colorado College Fedora 3.8 Utilities

See:
https://github.com/Tutt-Library/fedora38-utilities
"""
from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as fo:
    long_description = fo.read()

with open(path.join(here, 'VERSION'), encoding='utf-8') as fo:
    version = fo.read()

setup(
    name='Fedora3.8_Utilities',
    version=version,
    description='Utilities for legacy Fedora 3.8 repository',
    long_description = long_description, 
    url='https://github.com/Tutt-Library/fedora38-utilities',
    author='Jeremy Nelson',
    author_email='jermnelson@gmail.com',
    license='Apache2',
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Other Audience',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python', 
        'Programming Language :: Python :: 3.5',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    keywords='fedora3.8 library',
    packages=find_packages(exclude=("instance",)),
    install_requires=['elasticsearch',
                      'Flask',
                      'Flask-WTF',
                      'rdflib',
                      'requests'] 
)

    
